from typing import Union, Any, Optional, Dict
import logging
import sys
import time
import traceback

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.data import compute_data_stats
from graphsignal.span_context import start_span, stop_span, is_current_span

logger = logging.getLogger('graphsignal')

SAMPLE_TRACES = {1, 10, 100, 1000}


class TraceOptions:
    __slots__ = [
        'auto_sampling',
        'outlier_sampling',
        'ensure_sample',
        'enable_profiling'
    ]

    def __init__(self, 
            auto_sampling: bool = True,
            outlier_sampling: bool = True, 
            ensure_sample: bool = False, 
            enable_profiling: bool = False):
        self.auto_sampling = auto_sampling
        self.outlier_sampling = outlier_sampling
        self.ensure_sample = ensure_sample
        self.enable_profiling = enable_profiling

DEFAULT_OPTIONS = TraceOptions()


class DataObject:
    __slots__ = [
        'name',
        'obj',
        'counts',
        'check_missing_values'
    ]

    def __init__(self, name, obj, counts=None, check_missing_values=False):
        self.name = name
        self.obj = obj
        self.counts = counts
        self.check_missing_values = check_missing_values


class EndpointTrace:
    MAX_RUN_TAGS = 10
    MAX_TRACE_TAGS = 10
    MAX_RUN_PARAMS = 10
    MAX_TRACE_PARAMS = 10
    MAX_DATA_OBJECTS = 10

    __slots__ = [
        '_trace_sampler',
        '_metric_store',
        '_mv_detector',
        '_agent',
        '_endpoint',
        '_tags',
        '_params',
        '_options',
        '_is_sampling',
        '_latency_us',
        '_context',
        '_signal',
        '_is_stopped',
        '_start_counter',
        '_stop_counter',
        '_exc_info',
        '_data_objects',
        '_root_span',
        '_is_latency_outlier',
        '_has_missing_values'
    ]

    def __init__(self, endpoint, tags=None, options=None):
        if not endpoint:
            raise ValueError('endpoint is required')
        if not isinstance(endpoint, str):
            raise ValueError('endpoint must be string')
        if len(endpoint) > 50:
            raise ValueError('endpoint is too long (>50)')
        if tags is not None:
            if not isinstance(tags, dict):
                raise ValueError('tags must be dict')
            if len(tags) > EndpointTrace.MAX_TRACE_TAGS:
                raise ValueError('too many tags (>{0})'.format(EndpointTrace.MAX_TRACE_TAGS))

        self._endpoint = endpoint
        if tags is not None:
            self._tags = dict(tags)
        else:
            self._tags = None
        self._params = None
        if options is None:
            self._options = DEFAULT_OPTIONS
        else:
            self._options = options
        self._agent = graphsignal._agent
        self._trace_sampler = None
        self._metric_store = None
        self._mv_detector = None
        self._is_stopped = False
        self._is_sampling = False
        self._latency_us = None
        self._start_counter = None
        self._stop_counter = None
        self._context = False
        self._signal = None
        self._exc_info = None
        self._data_objects = None
        self._root_span = None
        self._is_latency_outlier = False
        self._has_missing_values = False

        try:
            self._start()
        except Exception as ex:
            if self._is_sampling:
                self._is_stopped = True
                self._trace_sampler.unlock()
            raise ex

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        if exc_info and exc_info[1] and isinstance(exc_info[1], Exception):
            self._exc_info = exc_info
        self.stop()
        return False

    def _init_sampling(self):
        self._is_sampling = True
        self._signal = self._agent.create_signal()
        self._context = {}

    def _start(self):
        if self._is_stopped:
            return

        self._trace_sampler = self._agent.trace_sampler(self._endpoint)
        self._metric_store = self._agent.metric_store(self._endpoint)
        self._mv_detector = self._agent.mv_detector()

        lock_group = 'samples'
        if self._options.auto_sampling:
            lock_group += '-auto'
        if self._options.ensure_sample:
            lock_group += '-ensured'
        if self._options.enable_profiling:
            lock_group += '-profiled'

        if ((self._options.auto_sampling and self._trace_sampler.lock(lock_group, include_trace_idx=SAMPLE_TRACES)) or
                (self._options.ensure_sample and self._trace_sampler.lock(lock_group, limit_per_interval=2))):
            self._init_sampling()

            # emit start event
            try:
                self._agent.emit_trace_start(self._signal, self._context, self._options)
            except Exception as exc:
                logger.error('Error in trace start event handlers', exc_info=True)
                self._add_agent_exception(exc)

        self._start_counter = time.perf_counter_ns()

        # stop current span
        self._root_span = start_span(name=self._endpoint, start_ns=self._start_counter, is_endpoint=True)

    def _measure(self) -> None:
        self._stop_counter = time.perf_counter_ns()

    def _stop(self) -> None:
        if self._is_stopped:
            return
        self._is_stopped = True

        if self._stop_counter is None:
            self._measure()
        self._latency_us = int((self._stop_counter - self._start_counter) / 1e3)
        end_us = _timestamp_us()

        # stop current span
        if is_current_span(self._root_span):
            stop_span(end_ns=self._stop_counter, has_exception=bool(self._exc_info and self._exc_info[0]))
        else:
            logger.error('Root span is not current span for endpoint {0}'.format(self._endpoint))

        if self._is_sampling:
            # emit stop event
            try:
                self._agent.emit_trace_stop(self._signal, self._context, self._options)
            except Exception as exc:
                logger.error('Error in trace stop event handlers', exc_info=True)
                self._add_agent_exception(exc)

        # if exception, but the trace is not being recorded, try to start tracing
        if not self._is_sampling and self._exc_info and self._exc_info[0]:
            if self._trace_sampler.lock('exceptions'):
                self._init_sampling()

        # update latency and counters
        if self._options.outlier_sampling:
            self._is_latency_outlier = self._metric_store.is_latency_outlier(self._latency_us, end_us)
            if not self._is_sampling and self._is_latency_outlier:
                if self._trace_sampler.lock('latency-outliers'):
                    self._init_sampling()

        self._metric_store.add_latency(self._latency_us, end_us)
        self._metric_store.inc_call_count(1, end_us)

        # compute data statistics
        data_stats = None
        if self._data_objects is not None:
            data_stats = {}
            for data_obj in self._data_objects.values():
                try:
                    stats = compute_data_stats(data_obj.obj)
                    if not stats:
                        continue
                    data_stats[data_obj.name] = stats

                    # add/update extra counts to computed counts
                    if data_obj.counts is not None:
                        stats.counts.update(data_obj.counts)

                    # check missing values
                    if data_obj.check_missing_values:
                        if self._mv_detector.detect(data_obj.name, stats.counts):
                            self._has_missing_values = True

                    # update data metrics
                    for count_name, count in stats.counts.items():
                        self._metric_store.inc_data_counter(
                            data_obj.name, count_name, count, end_us)
                except Exception as exc:
                    logger.error('Error computing data stats', exc_info=True)
                    self._add_agent_exception(exc)

        # if missing values detected, but the trace is not being recorded, try to start tracing
        if not self._is_sampling and self._has_missing_values:
            if self._trace_sampler.lock('missing-values'):
                self._init_sampling()

        # update exception counter
        if self._exc_info and self._exc_info[0]:
            self._metric_store.inc_exception_count(1, end_us)

        # fill and upload signal
        if self._is_sampling:
            # emit read event
            try:
                self._agent.emit_trace_read(self._signal, self._context, self._options)
            except Exception as exc:
                logger.error('Error in trace read event handlers', exc_info=True)
                self._add_agent_exception(exc)

            # copy data to signal
            self._signal.endpoint_name = self._endpoint
            self._signal.start_us = end_us - self._latency_us
            self._signal.end_us = end_us
            if self._exc_info and self._exc_info[0]:
                self._signal.trace_type = signals_pb2.TraceType.EXCEPTION_TRACE
            elif self._is_latency_outlier:
                self._signal.trace_type = signals_pb2.TraceType.LATENCY_OUTLIER_TRACE
            elif self._has_missing_values:
                self._signal.trace_type = signals_pb2.TraceType.MISSING_VALUES_TRACE
            else:
                self._signal.trace_type = signals_pb2.TraceType.SAMPLE_TRACE
            self._signal.process_usage.start_ms = self._agent._process_start_ms

            # copy tags
            tags = None
            if self._agent.tags is not None:
                tags = self._agent.tags.copy()
                if self._tags is not None:
                    tags.update(self._tags)
            elif self._tags is not None:
                tags = self._tags
            if tags is not None:
                for key, value in tags.items():
                    tag = self._signal.tags.add()
                    tag.key = key[:50]
                    tag.value = str(value)[:50]
                    if self._tags and key in self._tags:
                        tag.is_trace_level = True

            # copy params
            params = None
            if self._agent.params is not None:
                params = self._agent.params.copy()
                if self._params is not None:
                    params.update(self._params)
            elif self._params is not None:
                params = self._params
            if params is not None:
                for name, value in params.items():
                    param = self._signal.params.add()
                    param.name = name[:50]
                    param.value = str(value)[:50]
                    if self._params and name in self._params:
                        param.is_trace_level = True

            # copy trace measurements
            self._signal.trace_info.latency_us = self._latency_us
            self._signal.trace_info.is_ensured = self._options.ensure_sample
            self._signal.trace_info.is_profiled = self._options.enable_profiling

            # copy exception
            if self._exc_info and self._exc_info[0]:
                exception = self._signal.exceptions.add()
                if self._exc_info[0] and hasattr(self._exc_info[0], '__name__'):
                    exception.exc_type = str(self._exc_info[0].__name__)
                if self._exc_info[1]:
                    exception.message = str(self._exc_info[1])
                if self._exc_info[2]:
                    frames = traceback.format_tb(self._exc_info[2])
                    if len(frames) > 0:
                        exception.stack_trace = ''.join(frames)

            # copy data counts
            if data_stats is not None:
                for name, stats in data_stats.items():
                    data_stats_proto = self._signal.data_profile.add()
                    data_stats_proto.data_name = name
                    if stats.type_name:
                        data_stats_proto.data_type = stats.type_name
                    if stats.shape:
                        data_stats_proto.shape[:] = stats.shape
                    for name, count in stats.counts.items():
                        if count > 0:
                            dc = data_stats_proto.counts.add()
                            dc.name = name
                            dc.count = count

            # copy spans
            if self._root_span:
                _convert_span_to_proto(self._signal.root_span, self._root_span)

            # copy metrics
            self._metric_store.export(self._signal, end_us)

            # queue trace signal for upload
            self._agent.uploader().upload_trace(self._signal)
            self._agent.tick()

    def measure(self) -> None:
        if not self._is_stopped:
            self._measure()

    def stop(self) -> None:
        try:
            self._stop()
        finally:
            if self._is_sampling:
                self._is_stopped = True
                self._trace_sampler.unlock()

    def is_sampling(self):
        return self._is_sampling

    def set_tag(self, key: str, value: str) -> None:
        if not key:
            raise ValueError('set_tag: key must be provided')
        if value is None:
            raise ValueError('set_tag: value must be provided')

        if self._tags is None:
            self._tags = {}

        if len(self._tags) > EndpointTrace.MAX_TRACE_TAGS:
            raise ValueError('set_tag: too many tags (>{0})'.format(EndpointTrace.MAX_TRACE_TAGS))

        self._tags[key] = value

    def set_param(self, name: str, value: str) -> None:
        if not name:
            raise ValueError('set_param: name must be provided')
        if value is None:
            raise ValueError('set_param: value must be provided')

        if self._params is None:
            self._params = {}

        if len(self._params) > EndpointTrace.MAX_TRACE_PARAMS:
            raise ValueError('set_param: too many params (>{0})'.format(EndpointTrace.MAX_TRACE_PARAMS))

        self._params[name] = value

    def set_exception(self, exc: Optional[Exception] = None, exc_info: Optional[bool] = None) -> None:
        if exc is not None and not isinstance(exc, Exception):
            raise ValueError('set_exception: exc must be instance of Exception')

        if exc_info is not None and not isinstance(exc_info, bool):
            raise ValueError('set_exception: exc_info must be bool')

        if exc:
            self._exc_info = (exc.__class__, str(exc), exc.__traceback__)
        elif exc_info == True:
            self._exc_info = sys.exc_info()

    def set_data(self, 
            name: str, 
            obj: Any, 
            counts: Optional[Dict[str, int]] = None, 
            check_missing_values: Optional[bool] = False) -> None:
        if self._data_objects is None:
            self._data_objects = {}

        if name and not isinstance(name, str):
            raise ValueError('set_data: name must be string')

        if counts and not isinstance(counts, dict):
            raise ValueError('append_data: name must be dict')

        if len(self._data_objects) > EndpointTrace.MAX_DATA_OBJECTS:
            raise ValueError('set_data: too many data objects (>{0})'.format(EndpointTrace.MAX_DATA_OBJECTS))

        self._data_objects[name] = DataObject(
            name=name, obj=obj, counts=counts, check_missing_values=check_missing_values)

    def append_data(self, 
            name: str, 
            obj: Any, 
            counts: Optional[Dict[str, int]] = None, 
            check_missing_values: Optional[bool] = False) -> None:
        if self._data_objects is None:
            self._data_objects = {}

        if name and not isinstance(name, str):
            raise ValueError('append_data: name must be string')

        if counts and not isinstance(counts, dict):
            raise ValueError('append_data: name must be dict')

        if len(self._data_objects) > EndpointTrace.MAX_DATA_OBJECTS:
            raise ValueError('append_data: too many data objects (>{0})'.format(EndpointTrace.MAX_DATA_OBJECTS))

        if name in self._data_objects:
            data_obj = self._data_objects[name]
            data_obj.obj += obj
            if counts:
                if data_obj.counts is None:
                    data_obj.counts = {}
                for count_name, value in counts.items():
                    if count_name in data_obj.counts:
                        data_obj.counts[count_name] += value
                    else:
                        data_obj.counts[count_name] = value
        else:
            self._data_objects[name] = DataObject(
                name=name, obj=obj, counts=counts, check_missing_values=check_missing_values)

    def get_latency_us(self):
        return self._latency_us

    def _add_agent_exception(self, exc):
        if not self._is_sampling:
            return

        agent_error = self._signal.agent_errors.add()
        agent_error.message = str(exc)
        if exc.__traceback__:
            frames = traceback.format_tb(exc.__traceback__)
            if len(frames) > 0:
                agent_error.stack_trace = ''.join(frames)

    def repr(self):
        return 'EndpointTrace({0})'.format(self._endpoint)


def _timestamp_us():
    return int(time.time() * 1e6)


def _convert_span_to_proto(proto, span):
    if span.end_ns is None:
        return
    proto.name = span.name
    proto.start_ns = span.start_ns
    proto.end_ns = span.end_ns
    proto.has_exception = span.has_exception
    proto.is_endpoint = span.is_endpoint
    if span.children is not None:
        for child in span.children:
            _convert_span_to_proto(proto.spans.add(), child)



import logging
import sys
import threading
import random
import time
import uuid
import hashlib
import importlib

import graphsignal
from graphsignal import version
from graphsignal.uploader import Uploader
from graphsignal.metric_store import MetricStore
from graphsignal.trace_sampler import TraceSampler
from graphsignal.proto import signals_pb2
from graphsignal.data.missing_value_detector import MissingValueDetector
from graphsignal.proto_utils import parse_semver

logger = logging.getLogger('graphsignal')


class Agent:
    def __init__(
            self, 
            api_key=None, 
            api_url=None, 
            deployment=None, 
            tags=None, 
            auto_instrument=True, 
            upload_on_shutdown=True, 
            debug_mode=False):
        self.worker_id = _uuid_sha1(size=12)
        self.api_key = api_key
        if api_url:
            self.api_url = api_url
        else:
            self.api_url = 'https://agent-api.graphsignal.com'
        self.deployment = deployment
        self.tags = tags
        self.params = None
        self.auto_instrument = auto_instrument
        self.upload_on_shutdown = upload_on_shutdown
        self.debug_mode = debug_mode

        self._uploader = None
        self._trace_samplers = None
        self._metric_store = None
        self._mv_detector = None
        self._recorders = None

        self._process_start_ms = int(time.time() * 1e3)

    def setup(self):
        self._uploader = Uploader()
        self._uploader.setup()
        self._trace_samplers = {}
        self._metric_store = {}
        self._recorders = {}

        # pre-initialize recorders to enable auto-instrumentation for packages imported before graphsignal.configure()
        # as a fallback, any other trace sample will try to initialize uninitialized supported recorders
        # in a worst case scenario, this will result in a first execution not being traced
        self.recorders()

    def shutdown(self):
        # Create snapshot signals to send final metrics.
        if self.upload_on_shutdown:
            for endpoint, store in self._metric_store.items():
                if store.is_updated:
                    # upload metrics here
                    pass

        for recorder in self._recorders.values():
            recorder.shutdown()

        self.upload(block=True)

        self._recorders = None
        self._trace_samplers = None
        self._metric_store = None
        self._mv_detector = None
        self._uploader = None

    def uploader(self):
        return self._uploader

    def recorders(self):
        recorder_specs = [
            ('graphsignal.recorders.cprofile_recorder', 'CProfileRecorder', None, ['torch', 'yappi']),
            ('graphsignal.recorders.yappi_recorder', 'YappiRecorder', 'yappi', 'torch'),
            ('graphsignal.recorders.kineto_recorder', 'KinetoRecorder', 'torch', None),
            ('graphsignal.recorders.process_recorder', 'ProcessRecorder', None, None),
            ('graphsignal.recorders.nvml_recorder', 'NVMLRecorder', None, None),
            ('graphsignal.recorders.pytorch_recorder', 'PyTorchRecorder', 'torch', None),
            ('graphsignal.recorders.tensorflow_recorder', 'TensorFlowRecorder', 'tensorflow', None),
            ('graphsignal.recorders.jax_recorder', 'JAXRecorder', 'jax', None),
            ('graphsignal.recorders.onnxruntime_recorder', 'ONNXRuntimeRecorder', 'onnxruntime', None),
            ('graphsignal.recorders.xgboost_recorder', 'XGBoostRecorder', 'xgboost', None),
            ('graphsignal.recorders.deepspeed_recorder', 'DeepSpeedRecorder', 'deepspeed', None),
            ('graphsignal.recorders.openai_recorder', 'OpenAIRecorder', 'openai', None),
            ('graphsignal.recorders.langchain_recorder', 'LangChainRecorder', 'langchain', None),
            ('graphsignal.recorders.banana_recorder', 'BananaRecorder', 'banana_dev', None),
        ]
        last_exc = None
        for module_name, class_name, include, exclude in recorder_specs:
            try:
                key = (module_name, class_name)

                if exclude:
                    exclude = [exclude] if isinstance(exclude, str) else exclude
                    is_excluded = False
                    for mod in exclude:
                        if _check_module(mod):
                            if key in self._recorders:
                                del self._recorders[key]
                            is_excluded = True
                            break
                    if is_excluded:
                        continue

                if key not in self._recorders:
                    if not include or _check_module(include):
                        module = importlib.import_module(module_name)
                        recorder_class = getattr(module, class_name)
                        recorder = recorder_class()
                        recorder.setup()
                        self._recorders[key] = recorder
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

        return self._recorders.values()

    def trace_sampler(self, endpoint):
        if endpoint in self._trace_samplers:
            return self._trace_samplers[endpoint]
        else:
            trace_sampler = self._trace_samplers[endpoint] = TraceSampler()
            return trace_sampler

    def reset_metric_store(self, endpoint):
        self._metric_store[endpoint] = MetricStore()

    def metric_store(self, endpoint):
        if endpoint not in self._metric_store:
            self.reset_metric_store(endpoint)

        return self._metric_store[endpoint]

    def mv_detector(self):
        if self._mv_detector is None:
            self._mv_detector = MissingValueDetector()
        return self._mv_detector

    def emit_trace_start(self, signal, context, options):
        last_exc = None
        for recorder in reversed(self.recorders()):
            try:
                recorder.on_trace_start(signal, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_trace_stop(self, signal, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_trace_stop(signal, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def emit_trace_read(self, signal, context, options):
        last_exc = None
        for recorder in self.recorders():
            try:
                recorder.on_trace_read(signal, context, options)
            except Exception as exc:
                last_exc = exc
        if last_exc:
            raise last_exc

    def create_signal(self):
        signal = signals_pb2.Trace()
        signal.worker_id = graphsignal._agent.worker_id
        signal.trace_id = _uuid_sha1(size=12)
        if self.deployment:
            signal.deployment_name = self.deployment
        signal.agent_info.agent_type = signals_pb2.AgentInfo.AgentType.PYTHON_AGENT
        parse_semver(signal.agent_info.version, version.__version__)
        return signal

    def upload(self, block=False):
        if block:
            self._uploader.flush()
        else:
            self._uploader.flush_in_thread()

    def tick(self, block=False):
        self.upload(block=False)


def _check_module(module_name):
    return module_name in sys.modules


def _sha1(text, size=-1):
    sha1_hash = hashlib.sha1()
    sha1_hash.update(text.encode('utf-8'))
    return sha1_hash.hexdigest()[0:size]


def _uuid_sha1(size=-1):
    return _sha1(str(uuid.uuid4()), size)



import logging
import sys
import threading
import time
from collections import OrderedDict
import statistics

import graphsignal
from graphsignal import version
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')


class MetricStore:
    MAX_RESERVOIR_SIZE = 250
    MIN_SAMPLES = 30
    MAX_INTERVAL_SEC = 600

    def __init__(self):
        self._update_lock = threading.Lock()
        self.is_updated = False
        self.latency_reservoir = []
        self.latency_mean = 0
        self.latency_stdev = 0
        self.call_count = None
        self.exception_count = None
        self.data_counters = {}

    def add_latency(self, latency_us, timestamp_us):
        with self._update_lock:
            if len(self.latency_reservoir) < MetricStore.MAX_RESERVOIR_SIZE:
                self.latency_reservoir.append(latency_us)
            else:
                rand = timestamp_us % MetricStore.MAX_RESERVOIR_SIZE
                self.latency_reservoir[rand] = latency_us

        self.is_updated = True

    def is_latency_outlier(self, latency_us, timestamp_us):
        # Update mean and stdev approx every 10 samples to control overhead
        if (len(self.latency_reservoir) > MetricStore.MIN_SAMPLES and 
                (self.latency_mean == 0 or int(timestamp_us % 10) == 0)):
            self.latency_mean = statistics.mean(self.latency_reservoir)
            self.latency_stdev = statistics.stdev(self.latency_reservoir)

        # Samples worth investigating
        if self.latency_mean > 0:
            return self.latency_mean and latency_us > self.latency_mean + 6 * self.latency_stdev
        return False

    def inc_call_count(self, value, timestamp_us):
        with self._update_lock:
            if not self.call_count:
                self.call_count = OrderedDict()
            self._inc_counter(self.call_count, value, timestamp_us)
        
        self.is_updated = True

    def inc_exception_count(self, value, timestamp_us):
        with self._update_lock:
            if not self.exception_count:
                self.exception_count = OrderedDict()
            self._inc_counter(self.exception_count, value, timestamp_us)

        self.is_updated = True

    def inc_data_counter(self, data_name, counter_name, value, timestamp_us):
        with self._update_lock:
            key = (data_name, counter_name)
            if key not in self.data_counters:
                counter = OrderedDict()
                self.data_counters[key] = counter
            else:
                counter = self.data_counters[key]
            self._inc_counter(counter, value, timestamp_us)

        self.is_updated = True

    def _inc_counter(self, counter, value, timestamp_us):
        bucket = int(timestamp_us / 1e6)
        start_bucket = bucket - MetricStore.MAX_INTERVAL_SEC + 1

        has_expired = False
        for current_bucket in list(counter.keys()):
            if current_bucket < start_bucket:
                del counter[current_bucket]
                has_expired = True
            else:
                break
        # Any bucket expiration means the process is running for more than MAX_INTERVAL_SEC
        # Add a start bucket to properly compute rates in the backend
        if has_expired and start_bucket not in counter:
            counter[start_bucket] = 0

        counter[bucket] = counter.get(bucket, 0) + value

    def export(self, signal, timestamp_us):
        if self.latency_reservoir and len(self.latency_reservoir) > 0:
            self._convert_reservoir_to_proto(self.latency_reservoir, signal.trace_metrics.latency_us, timestamp_us)
        if self.call_count and len(self.call_count) > 0:
            self._convert_counter_to_proto(self.call_count, signal.trace_metrics.call_count, timestamp_us)
        if self.exception_count and len(self.exception_count) > 0:
            self._convert_counter_to_proto(self.exception_count, signal.trace_metrics.exception_count, timestamp_us)
        for (data_name, counter_name), data_counter in self.data_counters.items():
            data_metric = signal.data_metrics.add()
            data_metric.data_name = data_name
            data_metric.metric_name = counter_name
            self._convert_counter_to_proto(data_counter, data_metric.metric, timestamp_us)
        self.is_updated = False

    def _convert_reservoir_to_proto(self, reservoir, metric, timestamp_us):
        metric.type = signals_pb2.Metric.MetricType.RESERVOIR_METRIC
        for value in reservoir:
            metric.reservoir.values.append(value)

    def _convert_counter_to_proto(self, counter, metric, timestamp_us):
        metric.type = signals_pb2.Metric.MetricType.GAUGE_METRIC
        interval_sec = int(timestamp_us / 1e6) - min(counter.keys()) + 1
        metric.gauge = sum(counter.values()) / interval_sec




import unittest
import logging
import sys
import os
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.uploader import Uploader

logger = logging.getLogger('graphsignal')

class AgentTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    def test_create_signal(self):
        signal = graphsignal._agent.create_signal()
        self.assertTrue(signal.agent_info.version.major > 0 or signal.agent_info.version.minor > 0)

    @patch.object(Uploader, 'upload_trace')
    def test_shutdown_upload(self, mocked_upload_trace):
        with graphsignal.start_trace('test', options=graphsignal.TraceOptions(auto_sampling=False)):
            pass
        graphsignal.shutdown()

        #signal = mocked_upload_trace.call_args[0][0]

        #self.assertEqual(signal.deployment_name, 'd1')
        #self.assertEqual(signal.endpoint_name, 'test')
        #self.assertEqual(signal.trace_type, signals_pb2.TraceType.SNAPSHOT_TRACE)



import unittest
import logging
import sys
import time
import numpy as np
from unittest.mock import patch, Mock

import graphsignal
from graphsignal.proto import signals_pb2
from graphsignal.endpoint_trace import EndpointTrace, DEFAULT_OPTIONS
from graphsignal.recorders.process_recorder import ProcessRecorder
from graphsignal.uploader import Uploader

from graphsignal.data.missing_value_detector import MissingValueDetector

logger = logging.getLogger('graphsignal')


class EndpointTraceTest(unittest.TestCase):
    def setUp(self):
        if len(logger.handlers) == 0:
            logger.addHandler(logging.StreamHandler(sys.stdout))
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            tags={'k1': 'v1', 'k2': 'v2'},
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch.object(ProcessRecorder, 'on_trace_stop')
    @patch.object(Uploader, 'upload_trace')
    @patch('time.time', return_value=1000)
    def test_start_stop(self, mocked_time, mocked_upload_trace, mocked_process_on_trace_stop):
        graphsignal.set_tag('k3', 'v3')
        graphsignal.log_param('p1', 'v1')

        for i in range(10):
            trace = EndpointTrace(
                endpoint='ep1',
                tags={'k3': 'v33', 'k4': 4.0})
            trace.set_tag('k5', 'v5')
            trace.set_param('p2', 'v2')
            trace.set_data('input', np.asarray([[1, 2],[3, 4]]), counts=dict(c1=1, c2=2))
            time.sleep(0.01)
            trace.stop()

        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.deployment_name, 'd1')
        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.trace_type, signals_pb2.TraceType.SAMPLE_TRACE)
        self.assertTrue(signal.process_usage.start_ms > 0)
        self.assertEqual(signal.trace_metrics.call_count.gauge, 10.0)
        self.assertEqual(len(signal.trace_metrics.latency_us.reservoir.values), 10)
        self.assertEqual(signal.data_metrics[0].data_name, 'input')
        self.assertEqual(signal.data_metrics[0].metric_name, 'element_count')
        self.assertEqual(signal.data_metrics[0].metric.gauge, 40.0)
        self.assertEqual(signal.data_metrics[-2].metric_name, 'c1')
        self.assertEqual(signal.data_metrics[-2].metric.gauge, 10.0)
        self.assertEqual(signal.data_metrics[-1].metric_name, 'c2')
        self.assertEqual(signal.data_metrics[-1].metric.gauge, 20.0)
        self.assertEqual(signal.tags[0].key, 'k1')
        self.assertEqual(signal.tags[0].value, 'v1')
        self.assertTrue(not signal.tags[0].is_trace_level)
        self.assertEqual(signal.tags[1].key, 'k2')
        self.assertEqual(signal.tags[1].value, 'v2')
        self.assertEqual(signal.tags[2].key, 'k3')
        self.assertEqual(signal.tags[2].value, 'v33')
        self.assertEqual(signal.tags[3].key, 'k4')
        self.assertEqual(signal.tags[3].value, '4.0')
        self.assertEqual(signal.tags[4].key, 'k5')
        self.assertEqual(signal.tags[4].value, 'v5')
        self.assertTrue(signal.tags[4].is_trace_level)
        self.assertEqual(signal.params[0].name, 'p1')
        self.assertEqual(signal.params[0].value, 'v1')
        self.assertEqual(signal.params[1].name, 'p2')
        self.assertEqual(signal.params[1].value, 'v2')
        self.assertTrue(signal.params[1].is_trace_level)
        self.assertTrue(signal.trace_info.latency_us > 0)
        self.assertEqual(signal.data_profile[0].data_name, 'input')
        self.assertEqual(signal.data_profile[0].shape, [2, 2])
        self.assertEqual(signal.data_profile[0].counts[-2].name, 'c1')
        self.assertEqual(signal.data_profile[0].counts[-2].count, 1)
        self.assertEqual(signal.data_profile[0].counts[-1].name, 'c2')
        self.assertEqual(signal.data_profile[0].counts[-1].count, 2)

    @patch.object(ProcessRecorder, 'on_trace_start')
    @patch.object(Uploader, 'upload_trace')
    def test_start_exception(self, mocked_upload_trace, mocked_process_on_trace_start):
        mocked_process_on_trace_start.side_effect = Exception('ex1')
        trace = EndpointTrace(endpoint='ep1')
        trace.stop()

        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.trace_type, signals_pb2.TraceType.SAMPLE_TRACE)
        self.assertEqual(signal.agent_errors[0].message, 'ex1')
        self.assertNotEqual(signal.agent_errors[0].stack_trace, '')

    @patch.object(ProcessRecorder, 'on_trace_stop')
    @patch.object(Uploader, 'upload_trace')
    def test_agent_exception(self, mocked_upload_trace, mocked_process_on_trace_stop):
        mocked_process_on_trace_stop.side_effect = Exception('ex1')
        trace = EndpointTrace(
            endpoint='ep1')
        trace.stop()

        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.agent_errors[0].message, 'ex1')
        self.assertNotEqual(signal.agent_errors[0].stack_trace, '')

    @patch.object(Uploader, 'upload_trace')
    def test_inference_exception(self, mocked_upload_trace):

        for _ in range(2):
            try:
                with EndpointTrace(endpoint='ep1'):
                    raise Exception('ex1')
            except Exception as ex:
                if str(ex) != 'ex1':
                    raise ex

        self.assertEqual(mocked_upload_trace.call_count, 2)
        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.trace_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.trace_type, signals_pb2.TraceType.EXCEPTION_TRACE)
        self.assertEqual(signal.trace_metrics.exception_count.gauge, 2.0)
        self.assertEqual(signal.exceptions[0].exc_type, 'Exception')
        self.assertEqual(signal.exceptions[0].message, 'ex1')
        self.assertNotEqual(signal.exceptions[0].stack_trace, '')

    @patch.object(Uploader, 'upload_trace')
    def test_set_exception(self, mocked_upload_trace):
        trace = EndpointTrace(endpoint='ep1')
        try:
            raise Exception('ex2')
        except Exception as ex:
            trace.set_exception(ex)
        trace.stop()

        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.trace_type, signals_pb2.TraceType.EXCEPTION_TRACE)
        self.assertEqual(signal.trace_metrics.exception_count.gauge, 1)
        self.assertEqual(signal.exceptions[0].exc_type, 'Exception')
        self.assertEqual(signal.exceptions[0].message, 'ex2')
        self.assertNotEqual(signal.exceptions[0].stack_trace, '')

    @patch.object(Uploader, 'upload_trace')
    def test_set_exception_true(self, mocked_upload_trace):
        trace = EndpointTrace(endpoint='ep1')
        try:
            raise Exception('ex2')
        except Exception as ex:
            trace.set_exception(exc_info=True)
        trace.stop()

        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.trace_type, signals_pb2.TraceType.EXCEPTION_TRACE)
        self.assertEqual(signal.trace_metrics.exception_count.gauge, 1)
        self.assertEqual(signal.exceptions[0].exc_type, 'Exception')
        self.assertEqual(signal.exceptions[0].message, 'ex2')
        self.assertNotEqual(signal.exceptions[0].stack_trace, '')

    @patch.object(Uploader, 'upload_trace')
    def test_outlier(self, mocked_upload_trace):
        for _ in range(500):
            with EndpointTrace(endpoint='ep1'):
                time.sleep(0.00001)
        with EndpointTrace(endpoint='ep1'):
            time.sleep(0.01)

        has_outliers = False
        for call_args in mocked_upload_trace.call_args_list:
            signal = call_args[0][0]
            if signal.trace_type == signals_pb2.TraceType.LATENCY_OUTLIER_TRACE:
                has_outliers = True
                break
        self.assertTrue(has_outliers)

    @patch.object(Uploader, 'upload_trace')
    def test_set_data(self, mocked_upload_trace):
        with EndpointTrace(endpoint='ep1') as trace:
            trace.set_data('d1', {'c1': 100, 'c2': None}, check_missing_values=True)

        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')
        self.assertTrue(signal.worker_id != '')
        self.assertTrue(signal.start_us > 0)
        self.assertTrue(signal.end_us > 0)
        self.assertEqual(signal.trace_type, signals_pb2.TraceType.MISSING_VALUES_TRACE)

    @patch.object(Uploader, 'upload_trace')
    def test_spans(self, mocked_upload_trace):
        trace = EndpointTrace(endpoint='ep1')
        trace2 = EndpointTrace(endpoint='ep2')
        trace3 = EndpointTrace(endpoint='ep3')
        trace3.stop()
        trace2.stop()
        trace4 = EndpointTrace(endpoint='ep4')
        trace4.stop()
        trace.stop()

        signal = mocked_upload_trace.call_args[0][0]

        self.assertEqual(signal.endpoint_name, 'ep1')

        self.assertEqual(signal.root_span.name, 'ep1')
        self.assertTrue(signal.root_span.start_ns > 0)
        self.assertTrue(signal.root_span.end_ns > 0)

        self.assertEqual(signal.root_span.spans[0].name, 'ep2')
        self.assertTrue(signal.root_span.spans[0].is_endpoint)
        self.assertTrue(signal.root_span.spans[0].start_ns > signal.root_span.start_ns)
        self.assertTrue(signal.root_span.spans[0].end_ns < signal.root_span.end_ns)

        self.assertEqual(signal.root_span.spans[0].spans[0].name, 'ep3')
        self.assertTrue(signal.root_span.spans[0].spans[0].is_endpoint)
        self.assertTrue(signal.root_span.spans[0].spans[0].start_ns > signal.root_span.spans[0].start_ns)
        self.assertTrue(signal.root_span.spans[0].spans[0].end_ns < signal.root_span.spans[0].end_ns)

        self.assertEqual(signal.root_span.spans[1].name, 'ep4')
        self.assertTrue(signal.root_span.spans[1].start_ns > signal.root_span.spans[0].end_ns)
        self.assertTrue(signal.root_span.spans[1].end_ns < signal.root_span.end_ns)



import unittest
import logging
import sys
import os
import time
from unittest.mock import patch, Mock
from google.protobuf.json_format import MessageToJson
import pprint
import random

import graphsignal
from graphsignal.proto import signals_pb2

logger = logging.getLogger('graphsignal')

class MetricStoreTest(unittest.TestCase):
    def setUp(self):
        graphsignal.configure(
            api_key='k1',
            deployment='d1',
            upload_on_shutdown=False,
            debug_mode=True)

    def tearDown(self):
        graphsignal.shutdown()

    @patch('time.time', return_value=1)
    def test_update(self, mocked_time):
        store = graphsignal._agent.metric_store('ep1')

        store.add_latency(20, 1000 * 1e6)
        store.add_latency(30, 1000 * 1e6)

        store.inc_call_count(1, 1000 * 1e6)
        store.inc_call_count(1, 1000 * 1e6)
        store.inc_call_count(1, 1001 * 1e6)
    
        store.inc_exception_count(1, 100 * 1e6)
        store.inc_exception_count(300, 1001 * 1e6)

        store.inc_data_counter('d1', 'c1', 1, 1000 * 1e6)
        store.inc_data_counter('d1', 'c1', 1, 1000 * 1e6)
        store.inc_data_counter('d1', 'c1', 1, 1001 * 1e6)
        store.inc_data_counter('d1', 'c2', 1, 1001 * 1e6)

        signal = signals_pb2.Trace()
        store.export(signal, 1001 * 1e6)

        #pp = pprint.PrettyPrinter()
        #pp.pprint(MessageToJson(signal))

        self.assertEqual(signal.trace_metrics.latency_us.reservoir.values, [20, 30])
        self.assertEqual(signal.trace_metrics.call_count.gauge, 1.5)
        self.assertEqual(signal.trace_metrics.exception_count.gauge, 0.5)
        self.assertEqual(signal.data_metrics[0].data_name, 'd1')
        self.assertEqual(signal.data_metrics[0].metric_name, 'c1')
        self.assertEqual(signal.data_metrics[0].metric.gauge, 1.5)
        self.assertEqual(signal.data_metrics[1].data_name, 'd1')
        self.assertEqual(signal.data_metrics[1].metric_name, 'c2')
        self.assertEqual(signal.data_metrics[1].metric.gauge, 1.0)

    def test_is_latency_outlier(self):
        store = graphsignal._agent.metric_store('ep1')

        for i in range(1000):
            val = random.randint(100, 200)
            now = int(time.time() / 1e6) + i
            store.add_latency(val, now)
            self.assertFalse(store.is_latency_outlier(val, now))
        self.assertTrue(store.is_latency_outlier(1000, now + 1))


import logging
import sys
import xgboost as xgb


import graphsignal
from graphsignal.recorders.base_recorder import BaseRecorder
from graphsignal.proto_utils import parse_semver
from graphsignal.proto import signals_pb2
from graphsignal.proto_utils import add_framework_param, add_driver

logger = logging.getLogger('graphsignal')

class XGBoostRecorder(BaseRecorder):
    def __init__(self):
        self._framework = None
        self._comm_info = None

    def setup(self):
        self._framework = signals_pb2.FrameworkInfo()
        self._framework.name = 'XGBoost'
        parse_semver(self._framework.version, xgb.__version__)

        for key, value in xgb.get_config().items():
            if isinstance(value, (bool, int, float, str)):
                add_framework_param(self._framework, key, value)

    def on_trace_start(self, signal, context, options):
        pass

    def on_trace_stop(self, signal, context, options):
        pass

    def on_trace_read(self, signal, context, options):
        if self._framework:
            signal.frameworks.append(self._framework)



import logging

logger = logging.getLogger('graphsignal')


MISSING_COUNTERS = {
    'null_count',
    'nan_count',
    'inf_count',
    'empty_count'
}

class MissingValueDetector:
    def __init__(self):
        pass

    def detect(self, data_name, counts):
        for counter_name in MISSING_COUNTERS:
            if counter_name in counts:
                if counts[counter_name] > 0:
                    return True

        return False
